/* vim:set et sts=4: */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_EEKBOARD
#include <eekboard/eekboard-client.h>
#endif  /* HAVE_EEKBOARD */

#include "virtkbd.h"

struct _IBusM17NVirtualKeyboard {
    GObject parent;
    IBusEngine *engine;
    gchar *keyboard_type;
    GSList *keyboards;

#ifdef HAVE_EEKBOARD
    EekboardClient *client;
    EekboardContext *context;
#endif  /* HAVE_EEKBOARD */
};

struct _IBusM17NVirtualKeyboardClass {
    GObjectClass parent_class;
};

typedef struct _IBusM17NVirtualKeyboardClass IBusM17NVirtualKeyboardClass;

#ifdef HAVE_EEKBOARD
#define _EEK_INIT() eek_init ()
#else
#define _EEK_INIT()
#endif  /* HAVE_EEKBOARD */

G_DEFINE_TYPE_WITH_CODE (IBusM17NVirtualKeyboard, ibus_m17n_virtual_keyboard,
                         G_TYPE_OBJECT, _EEK_INIT());

static void
ibus_m17n_virtual_keyboard_dispose (GObject *object)
{
    IBusM17NVirtualKeyboard *virtkbd = IBUS_M17N_VIRTUAL_KEYBOARD (object);

    if (virtkbd->engine) {
        g_object_unref (virtkbd->engine);
        virtkbd->engine = NULL;
    }
    if (virtkbd->keyboards) {
        GSList *next = virtkbd->keyboards->next;
        /* virtkbd->keyboards is a ring; break it before free */
        virtkbd->keyboards->next = NULL;
        g_slist_free (next);
    }

#ifdef HAVE_EEKBOARD
    if (virtkbd->context) {
        g_object_unref (virtkbd->context);
        virtkbd->context = NULL;
    }
    if (virtkbd->client) {
        g_object_unref (virtkbd->client);
        virtkbd->client = NULL;
    }
#endif  /* HAVE_EEKBOARD */

    G_OBJECT_CLASS (ibus_m17n_virtual_keyboard_parent_class)->dispose (object);
}

static void
ibus_m17n_virtual_keyboard_finalize (GObject *object)
{
    IBusM17NVirtualKeyboard *virtkbd = IBUS_M17N_VIRTUAL_KEYBOARD (object);

    g_free (virtkbd->keyboard_type);

    G_OBJECT_CLASS (ibus_m17n_virtual_keyboard_parent_class)->finalize (object);
}

static void
ibus_m17n_virtual_keyboard_class_init (IBusM17NVirtualKeyboardClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    gobject_class->dispose = ibus_m17n_virtual_keyboard_dispose;
    gobject_class->finalize = ibus_m17n_virtual_keyboard_finalize;
}

static void
ibus_m17n_virtual_keyboard_init (IBusM17NVirtualKeyboard *self)
{
}

#ifdef HAVE_EEKBOARD
static void
destroyed_cb (EekboardClient *client,
              IBusM17NVirtualKeyboard *virtkbd)
{
    if (virtkbd->context) {
        g_object_unref (virtkbd->context);
        virtkbd->context = NULL;
    }
    if (virtkbd->client) {
        g_object_unref (virtkbd->client);
        virtkbd->client = NULL;
    }
}

static void
key_pressed_cb (EekboardContext *context,
                gchar           *keyname,
                EekSymbol       *symbol,
                guint            modifiers,
                IBusM17NVirtualKeyboard *virtkbd)
{
    if (EEK_IS_TEXT (symbol)) {
        const gchar *string;
        IBusText *text;

        string = eek_text_get_text (EEK_TEXT (symbol));
        text = ibus_text_new_from_static_string (string);
        ibus_engine_commit_text (virtkbd->engine, text);
    } else if (EEK_IS_KEYSYM (symbol)) {
        guint keyval = eek_keysym_get_xkeysym (EEK_KEYSYM (symbol));
        ibus_engine_forward_key_event (virtkbd->engine,
                                       keyval,
                                       0,
                                       modifiers);
        ibus_engine_forward_key_event (virtkbd->engine,
                                       keyval,
                                       0,
                                       modifiers | IBUS_RELEASE_MASK);
    } else if (g_strcmp0 (eek_symbol_get_name (symbol),
                          "cycle-keyboard") == 0) {
        virtkbd->keyboards = g_slist_next (virtkbd->keyboards);
        eekboard_context_set_keyboard (virtkbd->context,
                                       GPOINTER_TO_UINT (virtkbd->keyboards->data),
                                       NULL);
    }
}

static void
init_client (IBusM17NVirtualKeyboard *virtkbd)
{
    GDBusConnection *connection;
    GError *error;
    gchar **strv, **p;

    error = NULL;
    connection = g_bus_get_sync (G_BUS_TYPE_SESSION, NULL, &error);
    if (connection == NULL) {
        g_printerr ("Can't connect to the session bus: %s\n",
                    error->message);
        return;
    }

    virtkbd->client = eekboard_client_new (connection, NULL);
    g_object_unref (connection);
    g_signal_connect (virtkbd->client, "destroyed",
                      G_CALLBACK (destroyed_cb), virtkbd);

    virtkbd->context = eekboard_client_create_context (virtkbd->client,
                                                       "ibus-m17n",
                                                       NULL);
    g_signal_connect (virtkbd->context, "key-pressed",
                      G_CALLBACK (key_pressed_cb), virtkbd);

    strv = g_strsplit (virtkbd->keyboard_type, ",", -1);
    for (p = strv; *p; p++) {
        guint keyboard;

        keyboard = eekboard_context_add_keyboard (virtkbd->context,
                                                  g_strstrip (*p),
                                                  NULL);
        virtkbd->keyboards = g_slist_prepend (virtkbd->keyboards,
                                              GUINT_TO_POINTER (keyboard));
    }
    virtkbd->keyboards = g_slist_reverse (virtkbd->keyboards);
    g_strfreev (strv);

    /* make a cycle */
    g_slist_last (virtkbd->keyboards)->next = virtkbd->keyboards;

    eekboard_context_set_keyboard (virtkbd->context,
                                   GPOINTER_TO_UINT (virtkbd->keyboards->data),
                                   NULL);
}
#endif  /* HAVE_EEKBOARD */

void
ibus_m17n_virtual_keyboard_toggle_display (IBusM17NVirtualKeyboard *virtkbd)
{
#ifdef HAVE_EEKBOARD
    if (virtkbd->client == NULL)
        ibus_m17n_virtual_keyboard_enable (virtkbd);

    if (virtkbd->client != NULL) {
        if (eekboard_context_is_visible (virtkbd->context))
            eekboard_context_hide_keyboard (virtkbd->context, NULL);
        else
            eekboard_context_show_keyboard (virtkbd->context, NULL);
    }
#endif  /* HAVE_EEKBOARD */
}

void
ibus_m17n_virtual_keyboard_enable (IBusM17NVirtualKeyboard *virtkbd)
{
#ifdef HAVE_EEKBOARD
    if (virtkbd->client == NULL)
        init_client (virtkbd);

    if (virtkbd->client != NULL)
        eekboard_client_push_context (virtkbd->client, virtkbd->context, NULL);
#endif  /* HAVE_EEKBOARD */
}

void
ibus_m17n_virtual_keyboard_disable (IBusM17NVirtualKeyboard *virtkbd)
{
#ifdef HAVE_EEKBOARD
    if (virtkbd->client != NULL)
        eekboard_client_pop_context (virtkbd->client, NULL);
#endif  /* HAVE_EEKBOARD */
}

IBusM17NVirtualKeyboard *
ibus_m17n_virtual_keyboard_new (IBusEngine  *engine,
                                const gchar *keyboard_type)
{
    IBusM17NVirtualKeyboard *virtkbd = g_object_new (IBUS_M17N_TYPE_VIRTUAL_KEYBOARD, NULL);
    virtkbd->engine = g_object_ref (engine);
    virtkbd->keyboard_type = g_strdup (keyboard_type);
    return virtkbd;
}
